import os
from multiprocessing.pool import Pool
from pathlib import Path
from typing import List, Tuple
from urllib import request

import astropy.units as u
import numpy as np
import torch
from skimage.util import view_as_blocks
from sunpy.map import Map, make_fitswcs_header, all_coordinates_from_map

from itipy.data.dataset import SOHODataset, HMIContinuumDataset, STEREODataset, KSOFlatDataset, KSOFilmDataset
from itipy.data.editor import PaddingEditor, sdo_norms, hinode_norms, UnpaddingEditor


class InstrumentToInstrument:

    def __init__(self, model_name, model_path=None, device=None, depth_generator=3, patch_factor=0, n_workers=4):
        self.patch_factor = patch_factor
        self.depth_generator = depth_generator
        # Load Model
        if device is None:
            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        if model_path is None:
            model_path = self._getModelPath(model_name)
        self.generator = torch.load(model_path, map_location=device)
        self.generator.to(device)
        self.generator.eval()
        self.device = device
        self.n_workers = n_workers

    def translate(self, *args, **kwargs):
        raise NotImplementedError()

    def _translateDataset(self, dataset):
        with Pool(self.n_workers) as pool:
            for img, kwargs in pool.imap(dataset.convertData, dataset.data):
                #
                original_shape = img.shape
                img = np.array(img.data)  # remove np mask information
                #
                min_dim = min(
                    [i for i in range(img.shape[1], img.shape[1] * 2 ** (self.depth_generator + self.patch_factor))
                     if i % 2 ** (self.depth_generator + self.patch_factor) == 0])  # find min dim
                target_shape = (min_dim, min_dim)
                padding_editor = PaddingEditor(target_shape)
                # pad
                padded_img = padding_editor.call(img)
                padded_img = np.nan_to_num(padded_img, nan=np.nanmin(padded_img))
                # translate
                with torch.no_grad():
                    if self.patch_factor > 0:
                        iti_img = self._translateBlocks(padded_img, self.patch_factor)
                    else:
                        iti_img = self.generator(torch.tensor(padded_img).float().to(self.device).unsqueeze(0))
                        iti_img = iti_img[0].detach().cpu().numpy()
                # unpad
                scaling = iti_img.shape[-1] / padded_img.shape[-1]
                iti_img = UnpaddingEditor([p * scaling for p in original_shape[1:]]).call(iti_img)
                #
                ref_meta = [k['header'] for k in kwargs['kwargs_list']] if 'kwargs_list' in kwargs else [
                    kwargs['header']]
                # use last meta data as reference for additional observables
                ref_meta += [ref_meta[-1]] * (len(iti_img) - len(ref_meta))
                #
                ref_img = img.tolist()
                ref_img += [ref_img[-1]] * (len(iti_img) - len(ref_img))  # extend list
                ref_img = np.array(ref_img)
                #
                # create meta for additional channels
                maps = [Map(d, self._createMeta(d, ref_d, meta)) for d, ref_d, meta in zip(iti_img, ref_img, ref_meta)]
                maps = maps[0] if len(maps) == 1 else maps
                yield maps, img, iti_img

    def _createMeta(self, data, ref_data, ref_meta):
        scaling = data.shape[0] / ref_data.shape[0]
        ref_map = Map(ref_data, ref_meta)  # copy observer
        scale = (ref_meta['cdelt1'] / scaling, ref_meta['cdelt2'] / scaling)
        coord = ref_map.reference_coordinate
        wl = ref_map.wavelength if ref_map.waveunit is not None else None
        meta = make_fitswcs_header(data, coord, rotation_matrix=ref_map.rotation_matrix, scale=scale * u.arcsec / u.pix,
                                   observatory='ITI', instrument='ITI - ' + ref_map.instrument, wavelength=wl,
                                   exposure=1 * u.s, )
        return meta

    def _translateBlocks(self, img, n_patches):
        patch_dim = img.shape[-1] // n_patches
        #
        patch_shape = (img.shape[0], patch_dim, patch_dim)
        patches = view_as_blocks(img, patch_shape)
        patches = np.reshape(patches, (-1, *patch_shape))
        iti_patches = []
        with torch.no_grad():
            for patch in patches:
                iti_patch = self.generator(torch.tensor(patch).float().to(self.device).unsqueeze(0))
                iti_patches.append(iti_patch[0].detach().cpu().numpy())
        #
        iti_patches = np.array(iti_patches)
        iti_patches = iti_patches.reshape((n_patches, n_patches,
                                           iti_patches.shape[1], iti_patches.shape[2], iti_patches.shape[3]))
        iti_img = np.moveaxis(iti_patches, [0, 1], [1, 3]).reshape((iti_patches.shape[2],
                                                                    iti_patches.shape[0] * iti_patches.shape[3],
                                                                    iti_patches.shape[1] * iti_patches.shape[4]))
        #
        return iti_img

    def _getModelPath(self, model_name):
        model_path = os.path.join(Path.home(), '.iti', model_name)
        os.makedirs(os.path.join(Path.home(), '.iti'), exist_ok=True)
        if not os.path.exists(model_path):
            request.urlretrieve('http://kanzelhohe.uni-graz.at/iti/' + model_name, filename=model_path)
        return model_path

    def _adjustMeta(self, meta, new_data, scale_factor):
        # Update image scale and number of pixels
        new_meta = meta.copy()

        # Update metadata
        new_meta['cdelt1'] /= scale_factor
        new_meta['cdelt2'] /= scale_factor
        if 'CD1_1' in new_meta:
            new_meta['CD1_1'] /= scale_factor
            new_meta['CD2_1'] /= scale_factor
            new_meta['CD1_2'] /= scale_factor
            new_meta['CD2_2'] /= scale_factor
        new_meta['crpix1'] = (new_data.shape[1] + 1) / 2.
        new_meta['crpix2'] = (new_data.shape[0] + 1) / 2.
        s_map = Map(new_data, new_meta)
        lon, lat = s_map._get_lon_lat(s_map.center.frame)
        new_meta['crval1'] = lon.value
        new_meta['crval2'] = lat.value
        new_meta['naxis1'] = new_data.shape[1]
        new_meta['naxis2'] = new_data.shape[0]
        history = new_meta.get('history') + '; ' if 'history' in new_meta else ''
        new_meta['history'] = history + 'ITI enhanced'
        return new_meta


class InstrumentConverter:

    def _convertDataset(self, *datasets, n_workers=4) -> Tuple[np.ndarray, List]:
        images = []
        metas = []
        with Pool(n_workers) as pool:
            for data_sample in zip(*[pool.imap(ds.convertData, ds.data) for ds in datasets]):
                #
                images += [d for d, kwargs in data_sample]
                metas += [kwargs for d, kwargs in data_sample]
        return np.concatenate(images), metas


class SOHOToSDO(InstrumentToInstrument):

    def __init__(self, model_name='soho_to_sdo_v0_2.pt', **kwargs):
        super().__init__(model_name, **kwargs)
        self.norms = [sdo_norms[171], sdo_norms[193], sdo_norms[211], sdo_norms[304], sdo_norms['mag']]

    def translate(self, path, basenames=None, **kwargs):
        soho_dataset = SOHODataset(path, basenames=basenames, **kwargs)
        for maps, img, iti_img in self._translateDataset(soho_dataset):
            yield [Map(norm.inverse((s_map.data + 1) / 2), self.toSDOMeta(s_map.meta, instr))
                   for s_map, norm, instr in zip(maps, self.norms, ['AIA'] * 4 + ['HMI'])]

    def toSDOMeta(self, meta, instrument):
        wl_map = {171: 171, 195: 193, 284: 211, 304: 304, 6768: 6173, 0: 0}
        new_meta = meta.copy()
        new_meta['obsrvtry'] = 'SOHO-to-SDO'
        new_meta['telescop'] = 'sdo'
        new_meta['instrume'] = instrument
        new_meta['WAVELNTH'] = wl_map[meta.get('WAVELNTH', 0)]
        new_meta['waveunit'] = 'angstrom'
        return new_meta


class SOHOToSDOEUV(SOHOToSDO):

    def __init__(self, model_name='soho_to_sdo_euv_v0_1.pt', **kwargs):
        super().__init__(model_name, **kwargs)
        self.norms = [sdo_norms[171], sdo_norms[193], sdo_norms[211], sdo_norms[304]]

    def translate(self, path, basenames=None):
        soho_dataset = SOHODataset(path, basenames=basenames, wavelengths=[171, 195, 284, 304])
        for maps, img, iti_img in self._translateDataset(soho_dataset):
            yield [Map(norm.inverse((s_map.data + 1) / 2), self.toSDOMeta(s_map.meta, instr))
                   for s_map, norm, instr in zip(maps, self.norms, ['AIA'] * 4)]


class STEREOToSDO(InstrumentToInstrument):

    def __init__(self, model_name='stereo_to_sdo_v0_2.pt', **kwargs):
        super().__init__(model_name, **kwargs)

    def translate(self, path, basenames=None, return_arrays=False):
        soho_dataset = STEREODataset(path, basenames=basenames)
        for result, inputs, outputs in self._translateDataset(soho_dataset):
            norms = [sdo_norms[171], sdo_norms[193], sdo_norms[211], sdo_norms[304]]
            result = [Map(norm.inverse((s_map.data + 1) / 2), self.toSDOMeta(s_map.meta, instrument, wl))
                      for s_map, norm, instrument, wl in
                      zip(result, norms, ['AIA'] * 4, [171, 193, 211, 304])]
            if return_arrays:
                yield result, inputs, outputs
            else:
                yield result

    def toSDOMeta(self, meta, instrument, wl):
        new_meta = meta.copy()
        new_meta['obsrvtry'] = 'SOHO-to-SDO'
        new_meta['telescop'] = 'sdo'
        new_meta['instrume'] = instrument
        new_meta['WAVELNTH'] = wl
        new_meta['waveunit'] = 'angstrom'
        return new_meta


class STEREOToSDOMagnetogram(InstrumentToInstrument):

    def __init__(self, model_name='stereo_to_sdo_mag_v0_2.pt', **kwargs):
        super().__init__(model_name, **kwargs)

    def translate(self, path, basenames=None, return_arrays=False):
        soho_dataset = STEREODataset(path, basenames=basenames)
        for result, inputs, outputs in self._translateDataset(soho_dataset):
            norms = [sdo_norms[171], sdo_norms[193], sdo_norms[211], sdo_norms[304]]
            result = [Map(norm.inverse((s_map.data + 1) / 2), self.toSDOMeta(s_map.meta, 'AIA', wl))
                      for s_map, norm, wl in zip(result[:-1], norms, [171, 193, 211, 304])] + \
                     [self._createMagnetogramMap(result[-1].data, result[-1].meta)]
            if return_arrays:
                yield result, inputs, outputs
            else:
                yield result

    def _createMagnetogramMap(self, data, meta):
        v_max = sdo_norms['mag'].vmax
        s_map = Map((data + 1) / 2 * v_max, self.toSDOMeta(meta, 'HMI', 6173))
        hpc_coords = all_coordinates_from_map(s_map)
        r = np.sqrt(hpc_coords.Tx ** 2 + hpc_coords.Ty ** 2) / s_map.rsun_obs
        s_map.data[r > 1] = np.nan
        return s_map

    def toSDOMeta(self, meta, instrument, wl):
        new_meta = meta.copy()
        new_meta['obsrvtry'] = 'SOHO-to-SDO'
        new_meta['telescop'] = 'sdo'
        new_meta['instrume'] = instrument
        new_meta['WAVELNTH'] = wl
        new_meta['waveunit'] = 'angstrom'
        return new_meta


class KSOLowToHigh(InstrumentToInstrument):
    def __init__(self, model_name='kso_low_to_high_v0_2.pt', resolution=512, **kwargs):
        super().__init__(model_name, **kwargs)
        self.resolution = resolution

    def translate(self, paths, return_arrays=False, **kwargs):
        ds = KSOFlatDataset(paths, self.resolution, **kwargs)
        for result, inputs, outputs in self._translateDataset(ds):
            if return_arrays:
                yield result, inputs, outputs
            else:
                yield result


class KSOFilmToCCD(InstrumentToInstrument):
    def __init__(self, model_name='kso_film_to_ccd_v0_1.pt', resolution=512, **kwargs):
        super().__init__(model_name, **kwargs)
        self.resolution = resolution

    def translate(self, paths, return_arrays=False):
        ds = KSOFilmDataset(paths, self.resolution)
        for result, inputs, outputs in self._translateDataset(ds):
            if return_arrays:
                yield result, inputs, outputs
            else:
                yield result


class HMIToHinode(InstrumentToInstrument):
    def __init__(self, model_name='hmi_to_hinode_v0_2.pt', **kwargs):
        super().__init__(model_name, **kwargs)

    def translate(self, paths):
        ds = HMIContinuumDataset(paths)
        for s_map, input, output in self._translateDataset(ds):
            norm = hinode_norms['continuum']
            s_map = Map(norm.inverse((s_map.data + 1) / 2), s_map.meta)
            yield s_map


class KSOFlatConverter(InstrumentConverter):

    def __init__(self, resolution, **kwargs):
        super().__init__(**kwargs)
        self.resolution = resolution

    def convert(self, paths):
        ds = KSOFlatDataset(paths, self.resolution)
        return self._convertDataset(ds)



import argparse
import logging
import os

import numpy as np
from sunpy.visualization.colormaps import cm

from itipy.data.dataset import SDODataset, StorageDataset, STEREODataset
from itipy.data.editor import RandomPatchEditor, LambdaEditor, BrightestPixelPatchEditor, BlockReduceEditor
from itipy.train.model import DiscriminatorMode
from itipy.trainer import Trainer

parser = argparse.ArgumentParser(description='Train STEREO-To-SDO translations with synthetic magnetograms')
parser.add_argument('--base_dir', type=str, help='path to the results directory.')

parser.add_argument('--sdo_path', type=str, help='path to the SDO data.')
parser.add_argument('--stereo_path', type=str, help='path to the STEREO data.')
parser.add_argument('--sdo_converted_path', type=str, help='path to store the converted SDO data.')
parser.add_argument('--stereo_converted_path', type=str, help='path to store the converted STEREO data.')

args = parser.parse_args()
base_dir = args.base_dir

stereo_path = args.stereo_path
stereo_converted_path = args.stereo_converted_path
sdo_path = args.sdo_path
sdo_converted_path = args.sdo_converted_path

prediction_dir = os.path.join(base_dir, 'prediction')
os.makedirs(prediction_dir, exist_ok=True)

logging.basicConfig(
    level=logging.INFO,
    handlers=[
        logging.FileHandler("{0}/{1}.log".format(base_dir, "info_log")),
        logging.StreamHandler()
    ])

# Init Model
trainer = Trainer(4, 5, upsampling=0, discriminator_mode=DiscriminatorMode.CHANNELS, lambda_diversity=0,
                  norm='in_rs_aff')
trainer.cuda()
start_it = trainer.resume(base_dir)


# Init Dataset
def absolute_mag(data, **kwargs):
    data[-1] = np.abs(data[-1]) * 2 - 1
    return data


abs_mag_editor = LambdaEditor(absolute_mag)
block_reduce_editor = BlockReduceEditor(block_size=(1, 2, 2))

test_months = [11, 12]
train_months = list(range(2, 10))

sdo_train = SDODataset(sdo_path, resolution=2048, patch_shape=(1024, 1024), months=train_months)
sdo_train = StorageDataset(sdo_train, sdo_converted_path,
                           ext_editors=[RandomPatchEditor((512, 512)), block_reduce_editor, abs_mag_editor])

stereo_train = STEREODataset(stereo_path, months=test_months)
stereo_train = StorageDataset(stereo_train, stereo_converted_path,
                              ext_editors=[BrightestPixelPatchEditor((512, 512)), RandomPatchEditor((256, 256))])

sdo_valid = SDODataset(sdo_path, resolution=2048)
sdo_valid.addEditor(block_reduce_editor)
sdo_valid.addEditor(abs_mag_editor)
stereo_valid = STEREODataset(stereo_path)

plot_settings_A = [
    {"cmap": cm.sdoaia171, "title": "SECCHI 171", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia193, "title": "SECCHI 195", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia211, "title": "SECCHI 284", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia304, "title": "SECCHI 304", 'vmin': -1, 'vmax': 1}
]
plot_settings_B = [
    {"cmap": cm.sdoaia171, "title": "AIA 171", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia193, "title": "AIA 193", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia211, "title": "AIA 211", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia304, "title": "AIA 304", 'vmin': -1, 'vmax': 1},
    {"cmap": "gray", "title": "HMI Magnetogram", 'vmin': -1, 'vmax': 1}
]

# Start training
trainer.startBasicTraining(base_dir, stereo_train, sdo_train, stereo_valid, sdo_valid,
                           plot_settings_A, plot_settings_B)



import argparse
import logging
import os

from itipy.callback import HistoryCallback, ProgressCallback, \
    SaveCallback
from itipy.data.dataset import StorageDataset, KSOFlatDataset
from itipy.data.editor import BrightestPixelPatchEditor
from itipy.trainer import Trainer

parser = argparse.ArgumentParser(description='Train mitigation of atmospheric degradations from KSO Halpha observations')
parser.add_argument('--base_dir', type=str, help='path to the results directory.')

parser.add_argument('--hq_path', type=str, help='path to the high-quality KSO data.')
parser.add_argument('--lq_path', type=str, help='path to the low-quality KSO data.')
parser.add_argument('--hq_converted_path', type=str, help='path to store the converted high-quality KSO data.')
parser.add_argument('--lq_converted_path', type=str, help='path to store the converted low-quality KSO data.')
parser.add_argument('--resolution', type=int, help='resolution of the images (default=1024).',
                    default=1024, required=False)

args = parser.parse_args()
base_dir = args.base_dir
resolution = args.resolution
low_path = args.lq_path
low_converted_path = args.lq_converted_path
high_path = args.hq_path
high_converted_path = args.hq_converted_path

prediction_dir = os.path.join(base_dir, 'prediction')
os.makedirs(prediction_dir, exist_ok=True)

logging.basicConfig(
    level=logging.INFO,
    handlers=[
        logging.FileHandler("{0}/{1}.log".format(base_dir, "info_log")),
        logging.StreamHandler()
    ])

# Init Model
trainer = Trainer(1, 1)
trainer.cuda()
start_it = trainer.resume(base_dir)

# Init Dataset

train_months = list(range(2, 10))
valid_months = [11, 12]

q1_dataset = KSOFlatDataset(high_path, resolution, months=train_months)
q1_storage = StorageDataset(q1_dataset, high_converted_path,
                            ext_editors=[BrightestPixelPatchEditor((256, 256), random_selection=0.8)])

q2_dataset = KSOFlatDataset(low_path, resolution, months=train_months)
q2_storage = StorageDataset(q2_dataset, low_converted_path,
                            ext_editors=[BrightestPixelPatchEditor((256, 256), random_selection=0.8)])

q1_valid_dataset = KSOFlatDataset(high_path, resolution, months=valid_months)
q1_valid_dataset = StorageDataset(q1_valid_dataset, high_converted_path)

q2_valid_dataset = KSOFlatDataset(low_path, resolution, months=valid_months)
q2_valid_dataset = StorageDataset(q2_valid_dataset, low_converted_path)

# Init Callbacks
history = HistoryCallback(trainer, base_dir)
progress = ProgressCallback(trainer)
save = SaveCallback(trainer, base_dir)

plot_settings_A = {"cmap": "gray", "title": "Quality 2", 'vmin': -1, 'vmax': 1}
plot_settings_B = {"cmap": "gray", "title": "Quality 1", 'vmin': -1, 'vmax': 1}

trainer.startBasicTraining(base_dir, q2_storage, q1_storage, q2_valid_dataset, q1_valid_dataset,
                           plot_settings_A, plot_settings_B, num_workers=8)



import setuptools
from setuptools import setup

setup(
    name='itipy',
    version='0.1',
    packages=setuptools.find_packages(),
    url='https://github.com/RobertJaro/InstrumentToInstrument',
    license='GNU GENERAL PUBLIC LICENSE',
    author='Robert Jarolim',
    author_email='',
    description='Package for translation between image domains of different astrophysical instruments.',
    install_requires=['torch>=1.8', 'sunpy>=2.0', 'scikit-image', 'scikit-learn', 'tqdm',
                      'numpy', 'matplotlib', 'astropy', 'aiapy', 'drms', 'pytorch_fid']
)



import argparse
import logging
import os
import sys

from sunpy.visualization.colormaps import cm

from itipy.data.dataset import SDODataset, SOHODataset, StorageDataset
from itipy.data.editor import RandomPatchEditor
from itipy.train.model import DiscriminatorMode
from itipy.trainer import Trainer

parser = argparse.ArgumentParser(description='Ablation study training SOHO-To-SDO translations')
parser.add_argument('--base_dir', type=str, help='path to the results directory.')

parser.add_argument('--sdo_path', type=str, help='path to the SDO data.')
parser.add_argument('--soho_path', type=str, help='path to the SOHO data.')
parser.add_argument('--sdo_converted_path', type=str, help='path to store the converted SDO data.')
parser.add_argument('--soho_converted_path', type=str, help='path to store the converted SOHO data.')

parser.add_argument('--n_filters', type=str, help='number of filters in the first layer.')
parser.add_argument('--depth', type=str, help='number of down-/up-sampling blocks.')
parser.add_argument('--use_skip_connections', action='store_true', help='use skip connections between down- and up-sampling blocks.')

args = parser.parse_args()

n_filters = args.n_filters
depth = args.depth
skip_connections = args.use_skip_connections

base_dir = "%s/f%d_d%d_s%s_singlechanneldisc" % \
           (args.base_dir, n_filters, depth, skip_connections)

sdo_path = args.sdo_path
sdo_converted_path = args.sdo_path_converted
soho_path = args.soho_path
soho_converted_path = args.soho_converted_path

prediction_dir = os.path.join(base_dir, 'prediction')
os.makedirs(prediction_dir, exist_ok=True)

log_iteration = 1000
logging.basicConfig(
    level=logging.INFO,
    handlers=[
        logging.FileHandler("{0}/{1}.log".format(base_dir, "info_log")),
        logging.StreamHandler()
    ])

# Init Model
trainer = Trainer(5, 5, upsampling=1, discriminator_mode=DiscriminatorMode.SINGLE,
                  lambda_diversity=0, norm='in_rs_aff', skip_connections=skip_connections, depth_generator=depth,
                  n_filters=n_filters)
trainer.cuda()

# Init Dataset
train_months = list(range(2, 10))
test_months = [11, 12]

sdo_train = SDODataset(sdo_path, patch_shape=(1024, 1024), resolution=2048, months=train_months,
                       years=list(range(2011, 2021)))
sdo_train = StorageDataset(sdo_train, sdo_converted_path,
                           ext_editors=[RandomPatchEditor((256, 256))])

soho_train = SOHODataset(soho_path, resolution=1024, months=train_months, years=list(range(1996, 2010)))
soho_train = StorageDataset(soho_train, soho_converted_path,
                            ext_editors=[RandomPatchEditor((128, 128))])

sdo_valid = SDODataset(sdo_path, patch_shape=(1024, 1024), resolution=2048, limit=100, months=test_months)
sdo_valid = StorageDataset(sdo_valid, sdo_converted_path, ext_editors=[RandomPatchEditor((256, 256))])

soho_valid = SOHODataset(soho_path, resolution=1024, months=test_months, limit=100)
soho_valid = StorageDataset(soho_valid, soho_converted_path, ext_editors=[RandomPatchEditor((128, 128))])

plot_settings_A = [
    {"cmap": cm.sdoaia171, "title": "EIT 171", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia193, "title": "EIT 195", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia211, "title": "EIT 284", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia304, "title": "EIT 304", 'vmin': -1, 'vmax': 1},
    {"cmap": "gray", "title": "MDI Magnetogram", 'vmin': -1, 'vmax': 1}
]
plot_settings_B = [
    {"cmap": cm.sdoaia171, "title": "AIA 171", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia193, "title": "AIA 193", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia211, "title": "AIA 211", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia304, "title": "AIA 304", 'vmin': -1, 'vmax': 1},
    {"cmap": "gray", "title": "HMI Magnetogram", 'vmin': -1, 'vmax': 1},
]

# Start training
trainer.startBasicTraining(base_dir, soho_train, sdo_train, soho_valid, sdo_valid, plot_settings_A, plot_settings_B)



import argparse
import logging
import os

from sunpy.visualization.colormaps import cm

from itipy.data.dataset import SDODataset, StorageDataset
from itipy.data.editor import SliceEditor
from itipy.train.model import DiscriminatorMode
from itipy.trainer import Trainer

parser = argparse.ArgumentParser(description='Train SDO EUV-To-magnetogram translations (signed)')
parser.add_argument('--base_dir', type=str, help='path to the results directory.')

parser.add_argument('--sdo_path', type=str, help='path to the SDO data.')
parser.add_argument('--sdo_converted_path', type=str, help='path to store the converted SDO data.')

args = parser.parse_args()
base_dir = args.base_dir

sdo_path = args.sdo_path
sdo_converted_path = args.sdo_converted_path
prediction_dir = os.path.join(base_dir, 'prediction')
os.makedirs(prediction_dir, exist_ok=True)

logging.basicConfig(
    level=logging.INFO,
    handlers=[
        logging.FileHandler("{0}/{1}.log".format(base_dir, "info_log")),
        logging.StreamHandler()
    ])

# Init Model
trainer = Trainer(4, 5, upsampling=0, discriminator_mode=DiscriminatorMode.CHANNELS, lambda_diversity=0,
                  norm='in_aff', n_filters=32)
trainer.cuda()

sdo_train = SDODataset(sdo_path, resolution=512, months=list(range(2, 10)), years=list(range(2011, 2021)))
sdo_euv_train = StorageDataset(sdo_train, sdo_converted_path,
                               ext_editors=[SliceEditor(0, 4)])
sdo_mag_train = StorageDataset(sdo_train, sdo_converted_path, ext_editors=[])

sdo_valid = SDODataset(sdo_path, resolution=512, limit=100, months=[11, 12])
sdo_euv_valid = StorageDataset(sdo_valid, sdo_converted_path,
                               ext_editors=[SliceEditor(0, 4), ])
sdo_mag_valid = StorageDataset(sdo_valid, sdo_converted_path, ext_editors=[])

plot_settings_A = [
    {"cmap": cm.sdoaia171, "title": "SECCHI 171", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia193, "title": "SECCHI 195", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia211, "title": "SECCHI 284", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia304, "title": "SECCHI 304", 'vmin': -1, 'vmax': 1}
]
plot_settings_B = [
    {"cmap": cm.sdoaia171, "title": "AIA 171", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia193, "title": "AIA 193", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia211, "title": "AIA 211", 'vmin': -1, 'vmax': 1},
    {"cmap": cm.sdoaia304, "title": "AIA 304", 'vmin': -1, 'vmax': 1},
    {"cmap": "gray", "title": "HMI Magnetogram", 'vmin': -1, 'vmax': 1}
]

trainer.startBasicTraining(base_dir, sdo_euv_train, sdo_mag_train, sdo_euv_valid, sdo_mag_valid,
                           plot_settings_A, plot_settings_B, num_workers=4)



import logging
import os
from abc import ABC, abstractmethod

import numpy as np
import torch
from matplotlib import pyplot as plt
from torch.utils.data import DataLoader
from tqdm import tqdm

from itipy.trainer import Trainer


class Callback(ABC):

    def __init__(self, log_iteration=1000):
        self.log_iteration = log_iteration

    def __call__(self, iteration):
        if (iteration + 1) % self.log_iteration == 0:
            self.call(iteration)

    @abstractmethod
    def call(self, iteration, **kwargs):
        raise NotImplementedError()


class BasicPlot(Callback):

    def __init__(self, data, model: Trainer, path, plot_id, plot_settings, dpi=100, batch_size=None, **kwargs):
        self.data = data
        self.path = path
        self.model = model
        self.plot_settings = plot_settings
        self.dpi = dpi
        self.plot_id = plot_id
        self.batch_size = batch_size if batch_size is not None else len(data)

        super().__init__(**kwargs)

    def call(self, iteration, **kwargs):
        data = self.loadData()

        rows = len(data)
        columns = len(data[0])

        f, axarr = plt.subplots(rows, columns, figsize=(3 * columns, 3 * rows))
        axarr = np.reshape(axarr, (rows, columns))
        for i in range(rows):
            for j in range(columns):
                plot_settings = self.plot_settings[j].copy()
                ax = axarr[i, j]
                ax.axis("off")
                ax.set_title(plot_settings.pop("title", None))
                ax.imshow(data[i][j], **plot_settings)
        plt.tight_layout()
        path = os.path.join(self.path, "%s_iteration%06d.jpg" % (self.plot_id, iteration + 1))
        plt.savefig(path, dpi=self.dpi)
        plt.close()
        del f, axarr, data

    def loadData(self):
        with torch.no_grad():
            loader = DataLoader(self.data, batch_size=self.batch_size, shuffle=False)
            data, predictions = [], []
            for data_batch in loader:
                data_batch = data_batch.float().cuda()
                predictions_batch = self.predict(data_batch)
                data += [data_batch.detach().cpu().numpy()]
                predictions += [[pred.detach().cpu().numpy() for pred in predictions_batch]]
            data = np.concatenate(data)
            predictions = map(list, zip(*predictions)) # transpose
            predictions = [np.concatenate(p) for p in predictions]
            samples = [data, ] + [*predictions]
            # separate into rows and columns
            return [[d[j, i] for d in samples for i in range(d.shape[1])] for j in
                    range(len(self.data))]

    def predict(self, input_data):
        raise NotImplementedError()


class PlotABA(BasicPlot):
    def __init__(self, data, model, path, plot_settings_A=None, plot_settings_B=None, plot_id="ABA",
                 **kwargs):
        plot_settings_A = plot_settings_A if plot_settings_A is not None else [{"cmap": "gray"}] * model.input_dim_a
        plot_settings_A = plot_settings_A if isinstance(plot_settings_A, list) else [
                                                                                        plot_settings_A] * model.input_dim_a
        plot_settings_B = plot_settings_B if plot_settings_B is not None else [{"cmap": "gray"}] * model.input_dim_b
        plot_settings_B = plot_settings_B if isinstance(plot_settings_B, list) else [
                                                                                        plot_settings_B] * model.input_dim_b

        plot_settings = [*plot_settings_A, *plot_settings_B, *plot_settings_A]

        super().__init__(data, model, path, plot_id, plot_settings, **kwargs)

    def predict(self, x):
        x_ab, x_aba = self.model.forwardABA(x)
        return x_ab, x_aba


class PlotBAB(BasicPlot):
    def __init__(self, data, model, path, plot_settings_A=None, plot_settings_B=None, plot_id="BAB",
                 **kwargs):
        plot_settings_A = plot_settings_A if plot_settings_A is not None else [{"cmap": "gray"}] * model.input_dim_a
        plot_settings_A = plot_settings_A if isinstance(plot_settings_A, list) else [
                                                                                        plot_settings_A] * model.input_dim_a
        plot_settings_B = plot_settings_B if plot_settings_B is not None else [{"cmap": "gray"}] * model.input_dim_b
        plot_settings_B = plot_settings_B if isinstance(plot_settings_B, list) else [
                                                                                        plot_settings_B] * model.input_dim_b

        plot_settings_A = plot_settings_A if isinstance(plot_settings_A, list) \
            else [plot_settings_A] * model.input_dim_a
        plot_settings_B = plot_settings_B if isinstance(plot_settings_B, list) \
            else [plot_settings_B] * model.input_dim_b

        plot_settings = [*plot_settings_B, *plot_settings_A, *plot_settings_B]

        super().__init__(data, model, path, plot_id, plot_settings, **kwargs)

    def predict(self, x):
        x_ba, x_bab = self.model.forwardBAB(x)
        return x_ba, x_bab


class PlotAB(BasicPlot):
    def __init__(self, data, model, path, plot_settings_A=None, plot_settings_B=None, plot_id="AB",
                 **kwargs):
        plot_settings_A = plot_settings_A if plot_settings_A is not None else [{"cmap": "gray"}] * model.input_dim_a
        plot_settings_A = plot_settings_A if isinstance(plot_settings_A, list) else [
                                                                                        plot_settings_A] * model.input_dim_a
        plot_settings_B = plot_settings_B if plot_settings_B is not None else [{"cmap": "gray"}] * model.input_dim_b
        plot_settings_B = plot_settings_B if isinstance(plot_settings_B, list) else [
                                                                                        plot_settings_B] * model.input_dim_b

        plot_settings = [*plot_settings_A, *plot_settings_B]

        super().__init__(data, model, path, plot_id, plot_settings, **kwargs)

    def predict(self, input_data):
        x_ab = self.model.forwardAB(input_data)
        return (x_ab,)


class VariationPlotBA(BasicPlot):

    def __init__(self, data, model, path, n_samples, plot_settings_A=None, plot_settings_B=None, plot_id="variation",
                 **kwargs):
        self.n_samples = n_samples

        plot_settings_A = plot_settings_A if plot_settings_A is not None else [{"cmap": "gray"}] * model.input_dim_a
        plot_settings_A = plot_settings_A if isinstance(plot_settings_A, list) else [plot_settings_A] * model.input_dim_a
        plot_settings_A = plot_settings_A * n_samples
        plot_settings_B = plot_settings_B if plot_settings_B is not None else [{"cmap": "gray"}] * model.input_dim_b
        plot_settings_B = plot_settings_B if isinstance(plot_settings_B, list) else [
                                                                                        plot_settings_B] * model.input_dim_b

        plot_settings = [*plot_settings_B, *plot_settings_A]

        super().__init__(data, model, path, plot_id, plot_settings, **kwargs)

    def predict(self, x):
        x_ba = torch.cat([self.model.forwardBA(x) for _ in range(self.n_samples)], 1)
        return (x_ba,)


class HistoryCallback(Callback):
    def __init__(self, trainer: Trainer, path, log_iteration=1):
        self.trainer = trainer
        self.path = path

        self.loss = self.trainer.train_loss
        super().__init__(log_iteration)

    def call(self, iteration, **kwargs):
        self.loss['iteration'] += [iteration]
        self.loss['loss_gen_a_identity'] += [self.trainer.loss_gen_a_identity.cpu().detach().numpy()]
        self.loss['loss_gen_b_identity'] += [self.trainer.loss_gen_b_identity.cpu().detach().numpy()]
        self.loss['loss_gen_a_translate'] += [self.trainer.loss_gen_a_translate.cpu().detach().numpy()]
        self.loss['loss_gen_b_translate'] += [self.trainer.loss_gen_b_translate.cpu().detach().numpy()]
        self.loss['loss_gen_adv_a'] += [self.trainer.loss_gen_adv_a.cpu().detach().numpy()]
        self.loss['loss_gen_adv_b'] += [self.trainer.loss_gen_adv_b.cpu().detach().numpy()]
        self.loss['loss_gen_a_content'] += [self.trainer.loss_gen_a_content.cpu().detach().numpy()]
        self.loss['loss_gen_b_content'] += [self.trainer.loss_gen_b_content.cpu().detach().numpy()]
        self.loss['loss_gen_a_identity_content'] += [self.trainer.loss_gen_a_identity_content.cpu().detach().numpy()]
        self.loss['loss_gen_b_identity_content'] += [self.trainer.loss_gen_b_identity_content.cpu().detach().numpy()]
        self.loss['loss_gen_aba_noise'] += [self.trainer.loss_gen_aba_noise.cpu().detach().numpy()]
        self.loss['loss_gen_a_identity_noise'] += [self.trainer.loss_gen_a_identity_noise.cpu().detach().numpy()]
        self.loss['loss_dis_a'] += [self.trainer.loss_dis_a.cpu().detach().numpy()]
        self.loss['loss_dis_b'] += [self.trainer.loss_dis_b.cpu().detach().numpy()]
        self.loss['loss_gen_diversity'] += [self.trainer.loss_gen_diversity.cpu().detach().numpy()]
        if (iteration + 1) % 100 == 0:
            self.plotAdversarial()
            self.plotContent()
            self.plotDistortion()
            self.plotNoise()

    def plotAdversarial(self):
        plt.figure(figsize=(16, 8))
        plt.plot(self.loss['iteration'][25:-24], running_mean(self.loss['loss_gen_adv_a']), label='Generator A')
        plt.plot(self.loss['iteration'][25:-24], running_mean(self.loss['loss_gen_adv_b']), label='Generator B')
        plt.plot(self.loss['iteration'][25:-24], running_mean(self.loss['loss_dis_a']), label='Discriminator A')
        plt.plot(self.loss['iteration'][25:-24], running_mean(self.loss['loss_dis_b']), label='Discriminator B')
        plt.ylim((0, 0.9))
        plt.legend()
        plt.savefig(os.path.join(self.path, "progress_adversarial.jpg"), dpi=100)
        plt.close()

    def plotContent(self):
        plt.figure(figsize=(16, 8))
        plt.plot(self.loss['iteration'][25:-24], running_mean(self.loss['loss_gen_a_content']), label='Content A')
        plt.plot(self.loss['iteration'][25:-24], running_mean(self.loss['loss_gen_b_content']), label='Content B')
        plt.plot(self.loss['iteration'][25:-24], running_mean(self.loss['loss_gen_a_identity_content']), label='Content A Identity')
        plt.plot(self.loss['iteration'][25:-24], running_mean(self.loss['loss_gen_b_identity_content']), label='Content B Identity')
        plt.legend()
        plt.savefig(os.path.join(self.path, "progress_content.jpg"), dpi=100)
        plt.close()

    def plotNoise(self):
        plt.figure(figsize=(16, 8))
        plt.plot(self.loss['iteration'][25:-24], running_mean(self.loss['loss_gen_a_identity_noise']), label='Noise A Identity')
        plt.plot(self.loss['iteration'][25:-24], running_mean(self.loss['loss_gen_aba_noise']), label='Noise ABA')
        plt.plot(self.loss['iteration'][25:-24], running_mean(self.loss['loss_gen_diversity']), label='Diversity')
        plt.legend()
        plt.savefig(os.path.join(self.path, "progress_noise.jpg"), dpi=100)
        plt.close()

    def plotDistortion(self):
        plt.figure(figsize=(16, 8))
        plt.plot(self.loss['iteration'][25:-24], running_mean(self.loss['loss_gen_a_translate']), label='MAE A')
        plt.plot(self.loss['iteration'][25:-24], running_mean(self.loss['loss_gen_b_translate']), label='MAE B')
        plt.plot(self.loss['iteration'][25:-24], running_mean(self.loss['loss_gen_a_identity_content']), label='MAE A Identity')
        plt.plot(self.loss['iteration'][25:-24], running_mean(self.loss['loss_gen_b_identity_content']), label='MAE B Identity')
        plt.legend()
        plt.savefig(os.path.join(self.path, "progress_distortion.jpg"), dpi=100)
        plt.close()

class ValidationHistoryCallback(Callback):
    def __init__(self, trainer: Trainer, data_set_A, data_set_B, path, log_iteration=1000, num_workers=4):
        self.trainer = trainer
        self.path = path
        self.data_set_A = data_set_A
        self.data_set_B = data_set_B
        self.num_workers = num_workers

        self.loss = self.trainer.valid_loss
        super().__init__(log_iteration)

    def call(self, iteration, **kwargs):
        loss = {'loss_gen_a_identity': [],
                'loss_gen_b_identity': [],
                'loss_gen_a_translate': [],
                'loss_gen_b_translate': [],
                'loss_gen_adv_a': [],
                'loss_gen_adv_b': [],
                'loss_gen_a_content': [],
                'loss_gen_b_content': [],
                'loss_gen_a_identity_content': [],
                'loss_gen_b_identity_content': [],
                'loss_gen_aba_noise': [],
                'loss_gen_a_identity_noise': [],
                'loss_dis_a': [],
                'loss_dis_b': [],
                }
        dl_A, dl_B = DataLoader(self.data_set_A, batch_size=2, shuffle=False, num_workers=self.num_workers),\
                     DataLoader(self.data_set_B, batch_size=2, shuffle=False, num_workers=self.num_workers)
        for x_a, x_b in tqdm(zip(dl_A, dl_B), desc='Validation', total=len(dl_A)):
            x_a, x_b = x_a.float().cuda().detach(), x_b.float().cuda().detach()
            self.trainer.validate(x_a, x_b)
            loss['loss_gen_a_identity'] += [self.trainer.valid_loss_gen_a_identity.cpu().detach().numpy()]
            loss['loss_gen_b_identity'] += [self.trainer.valid_loss_gen_b_identity.cpu().detach().numpy()]
            loss['loss_gen_a_translate'] += [self.trainer.valid_loss_gen_a_translate.cpu().detach().numpy()]
            loss['loss_gen_b_translate'] += [self.trainer.valid_loss_gen_b_translate.cpu().detach().numpy()]
            loss['loss_gen_adv_a'] += [self.trainer.valid_loss_gen_adv_a.cpu().detach().numpy()]
            loss['loss_gen_adv_b'] += [self.trainer.valid_loss_gen_adv_b.cpu().detach().numpy()]
            loss['loss_gen_a_content'] += [self.trainer.valid_loss_gen_a_content.cpu().detach().numpy()]
            loss['loss_gen_b_content'] += [self.trainer.valid_loss_gen_b_content.cpu().detach().numpy()]
            loss['loss_gen_a_identity_content'] += [self.trainer.valid_loss_gen_a_identity_content.cpu().detach().numpy()]
            loss['loss_gen_b_identity_content'] += [self.trainer.valid_loss_gen_b_identity_content.cpu().detach().numpy()]
            loss['loss_gen_aba_noise'] += [self.trainer.valid_loss_gen_aba_noise.cpu().detach().numpy()]
            loss['loss_gen_a_identity_noise'] += [self.trainer.valid_loss_gen_a_identity_noise.cpu().detach().numpy()]
            loss['loss_dis_a'] += [self.trainer.valid_loss_dis_a.cpu().detach().numpy()]
            loss['loss_dis_b'] += [self.trainer.valid_loss_dis_b.cpu().detach().numpy()]

        self.loss['iteration'] += [iteration]
        self.loss['loss_gen_a_identity'].append(np.mean(loss['loss_gen_a_identity']))
        self.loss['loss_gen_b_identity'].append(np.mean(loss['loss_gen_b_identity']))
        self.loss['loss_gen_a_translate'].append(np.mean(loss['loss_gen_a_translate']))
        self.loss['loss_gen_b_translate'].append(np.mean(loss['loss_gen_b_translate']))
        self.loss['loss_gen_adv_a'].append(np.mean(loss['loss_gen_adv_a']))
        self.loss['loss_gen_adv_b'].append(np.mean(loss['loss_gen_adv_b']))
        self.loss['loss_gen_a_content'].append(np.mean(loss['loss_gen_a_content']))
        self.loss['loss_gen_b_content'].append(np.mean(loss['loss_gen_b_content']))
        self.loss['loss_gen_a_identity_content'].append(np.mean(loss['loss_gen_a_identity_content']))
        self.loss['loss_gen_b_identity_content'].append(np.mean(loss['loss_gen_b_identity_content']))
        self.loss['loss_gen_aba_noise'].append(np.mean(loss['loss_gen_aba_noise']))
        self.loss['loss_gen_a_identity_noise'].append(np.mean(loss['loss_gen_a_identity_noise']))
        self.loss['loss_dis_a'].append(np.mean(loss['loss_dis_a']))
        self.loss['loss_dis_b'].append(np.mean(loss['loss_dis_b']))


        self.plotAdversarial()
        self.plotContent()
        self.plotDistortion()
        self.plotNoise()

    def plotAdversarial(self):
        plt.figure(figsize=(16, 8))
        plt.plot(self.loss['iteration'], self.loss['loss_gen_adv_a'], label='Generator A')
        plt.plot(self.loss['iteration'], self.loss['loss_gen_adv_b'], label='Generator B')
        plt.plot(self.loss['iteration'], self.loss['loss_dis_a'], label='Discriminator A')
        plt.plot(self.loss['iteration'], self.loss['loss_dis_b'], label='Discriminator B')
        plt.legend()
        plt.savefig(os.path.join(self.path, "valid_progress_adversarial.jpg"), dpi=100)
        plt.close()

    def plotContent(self):
        plt.figure(figsize=(16, 8))
        plt.plot(self.loss['iteration'], self.loss['loss_gen_a_content'], label='Content A')
        plt.plot(self.loss['iteration'], self.loss['loss_gen_b_content'], label='Content B')
        plt.plot(self.loss['iteration'], self.loss['loss_gen_a_identity_content'], label='Content A Identity')
        plt.plot(self.loss['iteration'], self.loss['loss_gen_b_identity_content'], label='Content B Identity')
        plt.legend()
        plt.savefig(os.path.join(self.path, "valid_progress_content.jpg"), dpi=100)
        plt.close()

    def plotNoise(self):
        plt.figure(figsize=(16, 8))
        plt.plot(self.loss['iteration'], self.loss['loss_gen_a_identity_noise'], label='Noise A Identity')
        plt.plot(self.loss['iteration'], self.loss['loss_gen_aba_noise'], label='Noise ABA')
        plt.legend()
        plt.savefig(os.path.join(self.path, "valid_progress_noise.jpg"), dpi=100)
        plt.close()

    def plotDistortion(self):
        plt.figure(figsize=(16, 8))
        plt.plot(self.loss['iteration'], self.loss['loss_gen_a_translate'], label='MAE A')
        plt.plot(self.loss['iteration'], self.loss['loss_gen_b_translate'], label='MAE B')
        plt.plot(self.loss['iteration'], self.loss['loss_gen_a_identity_content'], label='MAE A Identity')
        plt.plot(self.loss['iteration'], self.loss['loss_gen_b_identity_content'], label='MAE B Identity')
        plt.legend()
        plt.savefig(os.path.join(self.path, "valid_progress_distortion.jpg"), dpi=100)
        plt.close()

class ProgressCallback(Callback):
    def __init__(self, trainer: Trainer, log_iteration=10):
        self.trainer = trainer
        super().__init__(log_iteration)

    def call(self, iteration, **kwargs):
        status = '[Iteration %08d] [D diff: %.03f/%.03f] [G loss: adv: %.03f/%.03f, recon: %.03f/%.03f, id: %.03f/%.03f, content: %.03f/%.03f]' % (
            iteration + 1,
            self.trainer.loss_dis_a, self.trainer.loss_dis_b,
            self.trainer.loss_gen_adv_a, self.trainer.loss_gen_adv_b,
            self.trainer.loss_gen_a_translate, self.trainer.loss_gen_b_translate,
            self.trainer.loss_gen_a_identity, self.trainer.loss_gen_b_identity,
            self.trainer.loss_gen_a_content, self.trainer.loss_gen_b_content,
        )
        logging.info('%s' % status)


class SaveCallback(Callback):
    def __init__(self, trainer: Trainer, checkpoint_dir, log_iteration=1000):
        self.trainer = trainer
        self.checkpoint_dir = checkpoint_dir
        super().__init__(log_iteration)

    def call(self, iteration, **kwargs):
        self.trainer.save(self.checkpoint_dir, iteration)


class NormScheduler(Callback):
    def __init__(self, trainer:Trainer, step=10000, gamma=0.5, init_iteration=0):
        self.trainer = trainer
        super(NormScheduler, self).__init__(step)
        self.momentum = 0.8
        self.gamma = gamma
        trainer.updateMomentum(self.momentum)
        [self.call(i) for i in range(init_iteration // step)]

    def call(self, iteration, **kwargs):
        if self.momentum == 0:
            return
        self.momentum *= self.gamma
        if self.momentum < 0.001:
            self.momentum = 0
        logging.info('Update normalization momentum: %.03f' % self.momentum)
        self.trainer.updateMomentum(self.momentum)

def running_mean(x, N=50):
    return np.convolve(x, np.ones((N,))/N, mode='valid')